/*
 * Copyright (C) 2010 August Mayer
 * All rights reserved.
 */

import boone.NetFactory;
import boone.NeuralNet;
import boone.PatternSet;
import boone.io.Proben1PatternFilter;
import evSOLve.JEvolution.JEvolution;
import evSOLve.JEvolution.JEvolutionException;
import evSOLve.JEvolution.chromosomes.Chromosome;
import evSOLve.JEvolution.chromosomes.RealChromosome;
import netevo.Evolver;
import netevo.decoders.BitDecoder;
import netevo.decoders.Decoder;
import netevo.decoders.RealDecoder;
import netevo.features.*;
import netevo.fitness.Classification;
import netevo.fitness.FitnessFunction;
import netevo.misc.NetReporter;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.List;

/**
 * Test Evolution of Neural Networks.
 *
 * @author August Mayer
 * @author Helmut A. Mayer
 * @version $Id: EvoTesterGlass.java 2293 2016-04-08 17:31:20Z helmut $
 */
public class EvoTesterGlass {

	public static String patternFile = "test/glass2.dt";

	// Note: Mutation rate is automatically set to 5/total_len.

	public int soupType = Chromosome.BIASED;

	public static double singleXRate = 0.817;

	public static double multiXRate = 0.1;

	public double shuffleRate = 0.5;

	public int sigmaStrategy = RealChromosome.EACH;

	public double minWeight = -5.0;

	public double maxWeight = 5.0;

	public PatternSet trainPatterns;

	public PatternSet testPatterns;

	/** The current run. */
	protected int run;


	public static void main(String[] args) throws Exception {

		EvoTesterGlass evoTester = new EvoTesterGlass();

		// First, load the patterns.
		PatternSet p1set = PatternSet.load(new File(patternFile), new Proben1PatternFilter());
		evoTester.trainPatterns = p1set.subSet(0, 161);        		// training set includes the validation patterns
		evoTester.testPatterns = p1set.subSet(161, 214);			// test set, pattern counts are in file

		// Create the neural network.
		NeuralNet net = NetFactory.createFeedForward(new int[]{9, 8, 6},
				true,    // fully connected
				null,    // Sigmoid activation function
				null,    // Rprop trainer
				null,    // Standard neuron
				null);    // Standard link

		// Perform evolution

		evoTester.test("glass-single", net,
				2  /*runs*/, 50 /*generations*/, 150 /*individuals*/,	// for testing 30 /*runs*/, 1000 /*generations*/, 150 /*individuals*/,   // a "real" run
				new Classification(evoTester.testPatterns), singleXRate, false, true);

//		evoTester.test("glass-multi-bit", net,
//				2  /*runs*/, 50 /*generations*/, 150 /*individuals*/,      			// for testing 30, 1000, 150,
//				new Classification(evoTester.testPatterns), multiXRate, false, false);
//
//		evoTester.test("glass-multi-real", net,
//				2  /*runs*/, 50 /*generations*/, 150 /*individuals*/,      			// for testing 30, 1000, 150,
//				new Classification(evoTester.testPatterns), multiXRate, true, false);
	}


	/** Test evolution. */
	public void test(String name, NeuralNet net, int numRuns, int numGenerations, int numIndividuals,
					 FitnessFunction fitFn, double crossoverRate, boolean type, boolean pack)
			throws JEvolutionException, IOException {

//		fitFn.setRegFactor(new RegularizationFactor(RegularizationFactor.REG_LINK));
//		Evolver evolver = new Evolver(net, fitFn);
		Evolver evolver = new Evolver(net, trainPatterns, fitFn);

		// Set up JEvolution.
		JEvolution evo = JEvolution.getInstance();
		evo.setPopulationSize(numIndividuals);
		evo.setMaximalGenerations(3);
//		evo.setMaximalGenerations(numGenerations);

		// Set up the specific parameters of this evolution variant.
		evo.setShuffleRate(shuffleRate);  // 0 = off

		// Create the feature instances.

		Decoder decoder = createDecoder(type);

		evolver.addFeature(new Links(decoder));
		if (!pack)           								// if packed, use the same decoder/chromosome
			decoder = createDecoder(type);
		evolver.addFeature(new Neurons(decoder));
		if (!pack)
			decoder = createDecoder(type);
		Feature weights = new Weights(decoder);
		weights.setMinValue(minWeight);
		weights.setMaxValue(maxWeight);
		evolver.addFeature(weights);
		if (!pack)
			decoder = createDecoder(type);
		Feature bp = new RpropTraining(decoder);
		bp.getField(BPTraining.EPOCHS).getDecoder().setGeneLength(10);	// 10 bits for epochs
		evolver.addFeature(bp);											// evolve BP parameters

		/* Sets up the reporter. */
		PrintWriter pw = new PrintWriter(System.out, true);
		NetReporter netReporter = new NetReporter(pw, name);
		JEvolution.getInstance().setReporter(netReporter);

		/* Sets up genotype and phenotype for evolution. */
		evolver.prepare();

		/* Determine and set the mutation rate. */
		List<Chromosome> genotype = evo.getGenotypeModel();
		int totalLen = 0;
		for (Chromosome c : genotype)
			totalLen += c.getLength();				// genotype length
		double mutationRate = 5.0 / totalLen;		// 5 is an educated guess
		for (Chromosome c : genotype) {
			c.setMutationRate(mutationRate);
			c.setCrossoverRate(crossoverRate);
		}

		System.out.println(evolver); 				// prints the evolution setup

		// Run JEvolution.
		for (run = 0; run < numRuns; run++) {
			System.out.println("Starting evolution run " + run + " for variant " + name + "\n");
			evo.doEvolve();
		}
		netReporter.reportEnd();
	}


	/** A utility to create a specific decoder with the given settings.
	 *
	 * @param real		if true, a RealDecoder, else a BitDecoder
	 *
	 * @return          a decoder
	 *
	 */
	private Decoder createDecoder(boolean real) {

		Decoder decoder;

		if (real) {
			decoder = new RealDecoder();
			RealChromosome rc = (RealChromosome)decoder.getChromosome();
			rc.setStrategy(sigmaStrategy);
		}
		else
			decoder = new BitDecoder();

		return decoder;
	}
}
