/*
 * Copyright (c) 2014.
 * August Mayer & Helmut A. Mayer
 * All rights reserved.
 */

import boone.NetFactory;
import boone.NeuralNet;
import boone.spike.SpikeSet;
import boone.spike.SpikingNeuralNet;
import boone.spike.SpikingNeuron;
import evSOLve.JEvolution.JEvolution;
import evSOLve.JEvolution.JEvolutionException;
import evSOLve.JEvolution.chromosomes.Chromosome;
import evSOLve.JEvolution.selection.TournamentSelection;
import netevo.Evolver;
import netevo.NetPhenotype;
import netevo.decoders.BitDecoder;
import netevo.decoders.Decoder;
import netevo.features.*;
import netevo.fitness.FitnessFunction;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * A sample netEvo application evolving a spiking neural network in such a way that the number of output spikes
 * (roughly) matches the number of input spikes.
 *
 * @author Helmut A. Mayer
 * @version $Id: TestEvolve.java 2293 2016-04-08 17:31:20Z helmut $
 * @since April 26, 2016
 */
public class SpikeApp {

	/** The starting point.
	 *
	 * @param args			arguments (unused)
	 * @throws Exception	just pass through potential exceptions
	 *
	 */
	public static void main(String[] args) throws Exception {

		new SpikeApp().evolve();
	}


	/** Evolve a spiking neural network.
	 *
	 * @throws java.io.IOException
	 * @throws evSOLve.JEvolution.JEvolutionException
	 */
	public void evolve() throws IOException, JEvolutionException {

		/* Create network and load patterns. */
		System.out.println("Creating test network.");

		SpikingNeuralNet net = NetFactory.createFeedForward(
				new int[]{2, 5, 2},      // 2 input, 5 hidden and 2 output neurons
				false,                   // layers are connected
				null,                    // default spike shape
				null,                    // default postsynaptic potential
				null,                    // no trainer
				null,                    // default neuron
				null);                   // default link



		/* Set up evolver and add features. */
		System.out.println("Setting up features to be evolved.");

		FitnessFunction ff = new SpikeCounter(net);
		Evolver evolver = new Evolver(net, ff);					// net evolution, no learning

		Decoder decoder = new BitDecoder();						// same decoder for all features = single chromosome
		evolver.addFeature(new Thresholds(decoder));
//		evolver.addFeature(new Delays(decoder));
//		evolver.addFeature(new AbsRefPeriods(decoder));
//		evolver.addFeature(new RelRefPeriods(decoder));
//		evolver.addFeature(new Links(decoder));
//		Neurons neurons = new Neurons(decoder);
//		neurons.setInputs(true);
//		evolver.addFeature(neurons);
//		evolver.getNetEvolver().setPruneInput(false);
//		evolver.addFeature(new Weights(decoder));

		/* Sets up genotype and phenotype for evolution. */
		evolver.prepare();

		/* Set up JEvolution. */
		System.out.println("Setting up evolution.");
		JEvolution evo = JEvolution.getInstance();
		evo.setMaximization(false);						// minimization problem
		evo.setFitnessThreshold(0.0);					// cannot get better
		evo.setPopulationSize(50);
		evo.setSelection(new TournamentSelection(2));	// keep it here!
		evo.setMaximalGenerations(100);

		/* Set the mutation rate. This has to be done after prepare(), as the latter
		 * determines the length of the chromosome(s). */
		Chromosome c = decoder.getChromosome();
		c.setMutationRate(3.0 / c.getLength());			// 3 is an educated guess
		c.setSoupType(Chromosome.BIASED);				// may improve start generation

		System.out.println(evolver); 					// prints the evolution setup

		/* Run JEvolution. */
		evo.doEvolve();

		/* Save the best network to a file. */
		NetPhenotype phenotype = (NetPhenotype)evo.getReporter().getBestIndividual().getPhenotype();
		NeuralNet bestNet = phenotype.getNet();
		System.out.println("Save best net.");
		bestNet.save(new File("best.xnet"));
	}


	/** A simple fitness function for a spiking neural network test. The desired goal is that the SNN generates a specific
	 * number of output spikes, which can be set via {@code targetCount}. The number of input spikes is set (more or less)
	 * to the number of desired output spikes, so generally, the SNN is fed n input spikes and should generate n output spikes.
	 * The fitness is the absolute difference of actual output spikes to desired output spikes. Thus, a fitness of 0 is optimal.
	 */
	class SpikeCounter extends FitnessFunction {

		/** The test set providing input spikes. */
		SpikeSet spikeSet;

		/** The desired number of output spikes. */
		private int targetCount;


		/** Constructs a spike counter. For the network test we need a spike set, which is constructed the following way:
		 *  input spikes are generated dependent on the target count of spikes, each input neuron receives its fair share
		 *  of spikes so that the number of total input spikes is (more or less) the number of target spikes. We also have
		 *  to link the input spikes to the respective input neurons. Internally, Boone uses the neuron ID to identify which
		 *  spikes belong to which neuron, so evolution keeps the IDs and everything works fine, but DO NOT change the IDs
		 *  in between, otherwise the network test will fail. The time interval of input spikes is [0, 1000)ms.
		 *
		 * @param net	an SNN
		 */
		SpikeCounter(NeuralNet net) {

			System.out.println("Creating a random input spike set.");
			spikeSet = new SpikeSet();
			targetCount = 13;

			for (int i = 0, n = net.getInputNeuronCount(); i < n; i++) {
				SpikingNeuron inputNeuron = (SpikingNeuron)net.getInputNeuron(i);			// for each input neuron
				List<Double> spikes = new ArrayList<>();
				for (int j = 0; j < targetCount / n; j++)                  					// input spikes
					spikes.add(Math.random() * 1000.0);                                     // within one second
				spikeSet.setInputSpikes(inputNeuron, spikes);
				System.out.printf("Input neuron with id %d receives spikes at: %s%n", inputNeuron.getID(), spikes);
			}
		}


		/** Runs the network with the input spikes and counts the number of generated output spikes. Returns the fitness as
		 * absolute difference of actual output spikes to the number of target spikes.
		 *
		 * @param net	an evolved SNN
		 *
		 * @return		the fitness (0 is optimal)
		 */
		public double getFitness(NeuralNet net) {

			runNet(net);													// let the network process all spikes
			int spikeCount = 0;

			for (int i = 0; i < net.getOutputNeuronCount(); i++) {
				SpikingNeuron outputNeuron = (SpikingNeuron)net.getOutputNeuron(i);
				List<Double> output = spikeSet.getOutputSpikes(outputNeuron);
//				System.out.printf("Spikes from output neuron with id %d: %s%n", outputNeuron.getID(), output);
				spikeCount += output.size();
			}

			return Math.abs(spikeCount - targetCount);
		}


		/** Feeds the input spikes, runs the network and waits until all spikes have been processed.
		 *
		 * @param net	the evolved SNN
		 */
		private void runNet(NeuralNet net) {

			SpikingNeuralNet snn = (SpikingNeuralNet)net;
			spikeSet.getOutputs().clear();											// remove all previous output

			for (int i = 0; i < snn.getOutputNeuronCount(); i++) {
				SpikingNeuron outputNeuron = (SpikingNeuron)snn.getOutputNeuron(i);
				outputNeuron.recordStart(spikeSet);                                 // record output into spike set
			}
			snn.setInput(spikeSet);                                                    // apply inputs
			snn.innervate();                                                        // start network processing
			while (snn.hasMoreSpikes()) {                                            // wait until all spikes are processed
				try {
					Thread.sleep(100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			snn.stop();                                    // stop processing
		}
	}

}
