/*
 * Copyright (c) 2014.
 * August Mayer & Helmut A. Mayer
 * All rights reserved.
 */

import boone.NetFactory;
import boone.NeuralNet;
import boone.PatternSet;
import boone.io.SNNSPatternFilter;
import evSOLve.JEvolution.JEvolution;
import evSOLve.JEvolution.JEvolutionException;
import evSOLve.JEvolution.chromosomes.Chromosome;
import evSOLve.JEvolution.selection.TournamentSelection;
import netevo.Evolver;
import netevo.NetPhenotype;
import netevo.decoders.BitDecoder;
import netevo.decoders.Decoder;
import netevo.features.Bias;
import netevo.features.Links;
import netevo.features.Neurons;
import netevo.features.Weights;
import netevo.fitness.Classification;
import netevo.fitness.FitnessFunction;

import java.io.File;
import java.io.IOException;

/**
 * A sample netEvo application evolving a network for the letters problem, where 26 different letters
 * are to be classified.
 *
 * @author August Mayer
 * @author Helmut A. Mayer
 * @version $Id: TestEvolve.java 2293 2016-04-08 17:31:20Z helmut $
 */
public class TestEvolve {

	/**
	 * The starting point.
	 *
	 * @param args arguments (unused)
	 *
	 * @throws Exception just pass through potential exceptions
	 */
	public static void main(String[] args) throws Exception {

		new TestEvolve().testLetters();
	}


	/**
	 * Evolve a network for the letters problem.
	 *
	 * @throws IOException
	 * @throws JEvolutionException
	 */
	public void testLetters() throws IOException, JEvolutionException {

		/* Create network and load patterns. */
		System.out.println("Creating test network.");

		NeuralNet net = NetFactory.createFeedForward(new int[]{35, 30, 26},
				true,    // fully connected
				null,    // Sigmoid activation function
				null,    // Rprop trainer
				null,    // Standard neuron
				null);    // Standard link

		System.out.println("Loading test patterns.");
		PatternSet pat = PatternSet.load(new File("test/letters.pat"), new SNNSPatternFilter());

		/* Set up evolver and add features. */
		System.out.println("Setting up features to be evolved.");

		FitnessFunction ff = new Classification(pat);
//		((Classification)ff).setAccelerate(true);
		Evolver evolver = new Evolver(net, ff);                // net evolution, no learning
//		Evolver evolver = new Evolver(net, pat, ff);		// net evolution with learning

		Decoder decoder = new BitDecoder();                    // same decoder for all features = single chromosome
		evolver.addFeature(new Links(decoder));
		Neurons neurons = new Neurons(decoder);
//		neurons.setInputs(true);
		evolver.addFeature(neurons);
//		evolver.getNetEvolver().setPruneInput(false);
		evolver.addFeature(new Weights(decoder));
		evolver.addFeature(new Bias(decoder));
//		evolver.addFeature(new Patterns(decoder));		// evolve training data set
//		Functions functions = new Functions(decoder);
//		functions.doOutputs(true);
//		functions.addFunction(new Function.Sigmoid(), null);
//		functions.addFunction(new Function.TanHyp(), null);
//		functions.addFunction(new Function.NaturalSpline(), evolver);
//		functions.addFunction(new Function.CatmullRomSpline(), evolver);
//		evolver.addFeature(functions);						// evolve activation function type

//		BPTraining bp = new BPTraining(decoder);
//		bp.getField(BPTraining.EPOCHS).getDecoder().setGeneLength(10);	// 10 bits for epochs
//		evolver.addFeature(bp);

		/* Sets up genotype and phenotype for evolution. */
		evolver.prepare();

		/* Set up JEvolution. */
		System.out.println("Setting up evolution.");
		JEvolution evo = JEvolution.getInstance();
		evo.setPopulationSize(200);
		evo.setSelection(new TournamentSelection(2));
		evo.setMaximalGenerations(10000);                // 10,000 and more for meaningful results

		/* Set the mutation rate. This has to be done after prepare(), as the latter
		 * determines the length of the chromosome(s). */
		Chromosome c = decoder.getChromosome();
		c.setMutationRate(3.0 / c.getLength());            // 3 is an educated guess
		c.setSoupType(Chromosome.BIASED);                // may improve start generation

		System.out.println(evolver);                    // prints the evolution setup

		/* Run JEvolution. */
		evo.doEvolve();

		/* Save the best network to a file. */
		NetPhenotype phenotype = (NetPhenotype)evo.getReporter().getBestIndividual().getPhenotype();
		NeuralNet bestNet = phenotype.getNet();
		System.out.println("Save best net.");
		bestNet.save(new File("best.xnet"));
	}
}
