/*
 * Copyright (C) 2010 August Mayer
 * All rights reserved.
 */

import boone.NetFactory;
import boone.NeuralNet;
import boone.PatternSet;
import boone.Trainer;
import boone.io.BooneFilter;
import boone.io.Proben1PatternFilter;
import boone.io.SNNSPatternFilter;
import boone.map.Function;
import boone.training.NetError;
import evSOLve.JEvolution.misc.Utilities;

import java.io.File;

/**
 * Test network training.
 * Note: This demonstrates the NetError class, which was originally part of NetEvo,
 * but is included into Boone itself now.
 *
 * @author August Mayer
 * @version $Id: TrainingTester.java 2293 2016-04-08 17:31:20Z helmut $
 */
public class TrainingTester {

	/**
	 * @param args the command line arguments
	 */
	public static void main(String[] args) throws Exception {

		runGlassTraining();
		//runTwoSpiralsTraining();
		//runMackeyGlassTraining();
	}


	protected static void runGlassTraining() throws Exception {

		PatternSet p1set = PatternSet.load(new File("test/glass2.dt"), new Proben1PatternFilter());
		System.out.println(p1set.size());
		PatternSet trainPatterns = p1set.subSet(0, 161);        	// training set includes the validation patterns
		PatternSet testPatterns = p1set.subSet(161, 214);			// test set, pattern counts are in file

		// Load the neural network.
		NeuralNet net = NetFactory.createFeedForward(new int[]{9, 8, 6},
				true,    // fully connected
				null,    // Sigmoid activation function
				null,    // Rprop trainer
				null,    // Standard neuron
				null);    // Standard link

		// Perform training.
		TrainingTester.test("Glass", "glass2 (train+val)", "9-8-6 fully-connected", net, trainPatterns,
				testPatterns, 10, 500, 4);
	}


	protected static void runTwoSpiralsTraining() throws Exception {

		PatternSet trainPatterns = PatternSet.load(new File("data/twoSpiral_trainval.pat"), new SNNSPatternFilter());
		PatternSet testPatterns = PatternSet.load(new File("data/twoSpiral_test.pat"), new SNNSPatternFilter());

		NeuralNet net = NetFactory.createFeedForward(new int[]{2, 18, 2},
				true,    // fully connected
				null,    // Sigmoid activation function
				null,    // Rprop trainer
				null,    // Standard neuron
				null);    // Standard link

		// Perform training.
		TrainingTester.test("TwoSpirals", "twoSpirals (train+val)", "2-18-2 fully-connected",
				net, trainPatterns, testPatterns, 10, 5000, 4);
	}


	protected static void runMackeyGlassTraining() throws Exception {

		PatternSet trainValPatterns = PatternSet.load(new File("data/mg84-trainval.xpat"), new BooneFilter("set", true));
		PatternSet testPatterns = PatternSet.load(new File("data/mg84-test.xpat"), new BooneFilter("set", true));

		NeuralNet net = NetFactory.createFeedForward(new int[]{4, 10, 10, 1},
				false,    // not fully connected
				null,    // Sigmoid activation function
				null,    // Rprop trainer
				null,    // Standard neuron
				null);    // Standard link

		// Set output neuron to linear
		net.getOutputNeuron(0).setActivationFn(new Function.Identity());

//		template.save(new File("mgnet.xnet"));

		// Perform training.
		TrainingTester.test("MackeyGlass", "mg84 (train+val)", "4-10-10-1 lin-out",
				net, trainValPatterns, testPatterns, 10, 500, 2);    // Note: NRMSE = 2
	}


	/**
	 * Test the network with training.
	 *
	 * @param name                 Textual description of the experiment.
	 * @param patternSetName       Textual description of the pattern set.
	 * @param neuralNetDescription Textual description of the neural network.
	 * @param net                  Neural network to train.
	 * @param trainPatterns        The patterns to use for training.
	 * @param testPatterns         The patterns to use for testing.
	 * @param numRuns              Number of runs, i.e. repetitions of the whole test.
	 * @param numEpochs            Number of epochs to train.
	 * @param whichError           Which error to use as the test error. See the NetError class.
	 */
	public static void test(String name, String patternSetName, String neuralNetDescription,
							NeuralNet net, PatternSet trainPatterns, PatternSet testPatterns,
							int numRuns, int numEpochs, int whichError) {

		final String[] ERROR_KIND = new String[]{"SSE", "MSE", "NRMSE", "SqE%", "Classification"};
		double[][] stats = new double[2][4];
		final int TRAIN = 0, TEST = 1;
		final int MEAN = 0, STD_DEV = 1, BEST = 2, WORST = 3;
		stats[TRAIN][BEST] = stats[TEST][BEST] = Double.MAX_VALUE;
		Trainer trainer = net.getTrainer();

		for (int run = 0; run < numRuns; run++) {
			net.randomize();
			trainer.setTrainingData(trainPatterns);
			trainer.setEpochs(numEpochs);
			trainer.train();

			double[] trainErrors = NetError.netErrors(net, trainPatterns, null, 0, 1);

			double trainErr = trainErrors[whichError];    // Classification error
			stats[TRAIN][MEAN] += trainErr;
			stats[TRAIN][STD_DEV] += trainErr * trainErr;
			stats[TRAIN][BEST] = Math.min(stats[TRAIN][BEST], trainErr);
			stats[TRAIN][WORST] = Math.max(stats[TRAIN][WORST], trainErr);

			double[] testErrors = NetError.netErrors(net, testPatterns, null, 0, 1);

			double testErr = testErrors[whichError];    // Classification error
			stats[TEST][MEAN] += testErr;
			stats[TEST][STD_DEV] += testErr * testErr;
			stats[TEST][BEST] = Math.min(stats[TEST][BEST], testErr);
			stats[TEST][WORST] = Math.max(stats[TEST][WORST], testErr);

			System.out.println("Run " + run +
					": Training set error = " + Utilities.fixedPoint.format(trainErr) +
					", test set error = " + Utilities.fixedPoint.format(testErr));

			System.out.println("  Training: " +
					trainPatterns.size() + " patterns, " + netErrorString(trainErrors));
			System.out.println("  Test: " +
					testPatterns.size() + " patterns, " + netErrorString(testErrors));
		}

		// Finish mean and stdDev calculation.
		stats[TRAIN][MEAN] /= numRuns;
		stats[TEST][MEAN] /= numRuns;
		stats[TRAIN][STD_DEV] =
				Math.sqrt((stats[TRAIN][STD_DEV] / numRuns) - stats[TRAIN][MEAN] * stats[TRAIN][MEAN]);
		stats[TEST][STD_DEV] =
				Math.sqrt((stats[TEST][STD_DEV] / numRuns) - stats[TEST][MEAN] * stats[TEST][MEAN]);

		// Report
		System.out.println();
		System.out.println(name + " Training Report:");
		System.out.println("- " + neuralNetDescription + " network");
		System.out.println("- " + patternSetName + " pattern set");
		System.out.println("- " + numRuns + " runs, " + numEpochs + " epochs");
		System.out.println("- " + ERROR_KIND[whichError] + " error");
		System.out.println("Pattern Set, Mean, StdDev, Best, Worst");
		for (int i = 0; i < 2; i++) {
			for (int j = -1; j < 4; j++)
				if (j == -1)
					System.out.print(i == 0 ? "Training, " : "Test, ");
				else
					System.out.print(Utilities.fixedPoint.format(stats[i][j]) + ", ");
			System.out.println();
		}
	}


	protected static String netErrorString(double[] errors) {

		return "SSE " + Utilities.fixedPoint.format(errors[0]) + ", MSE " + Utilities.fixedPoint.format(errors[1]) +
				", NRMSE " + Utilities.fixedPoint.format(errors[2]) + ", SqE% " +
				Utilities.fixedPoint.format(errors[3]) + ", ClErr " + Utilities.fixedPoint.format(errors[4]);
	}

}
